from abc import ABC, abstractmethod
from inspect import iscoroutinefunction
from inspect import getdoc
from importlib.machinery import SourceFileLoader
from pathlib import Path

class Fnc(ABC):

    registry = {}

    def __init__(self, name, help):
        if name in Fnc.registry:
            raise RuntimeError(f"Funcion '{name}' is already registered")
        if isinstance(help, str):
            help = help.strip()
            if not help:
                help = None
        self._help = help
        Fnc.registry[name] = self

    @abstractmethod
    async def __call__(self, env, **kwargs):
        ...

    @classmethod
    def search(cls, name):
        f = Fnc.registry.get(name)
        return f

    @property
    def help(self):
        return self._help


class PythonFnc(Fnc):

    def __init__(self, name, f, help=None):
        """
        врапер для функции python
        :param name: имя функции для вызова RPC
        :param f: ссылка на функцию или корутину
        :param help: документация. если не задана, то будет взята из комментария к функции
        """
        if not help:
            help = getdoc(f)
        if iscoroutinefunction(f):
            async def wrapper_env(env, **kwargs):
                nonlocal f
                return await f(env, **kwargs)
            async def wrapper(env, **kwargs):
                nonlocal f
                return await f(**kwargs)
        else:
            async def wrapper_env(env, **kwargs):
                nonlocal f
                return f(env, **kwargs)
            async def wrapper(env, **kwargs):
                nonlocal f
                return f(**kwargs)
        if "env" in f.__code__.co_varnames:
            self.__f = wrapper_env
        else:
            self.__f = wrapper
        super().__init__(name, help)

    async def __call__(self, env, **kwargs):
        return await self.__f(env, **kwargs)


class PluginFnc(PythonFnc):

    modules = {}

    def __init__(self, name, module_path, help=None, fnc_name=None, module_name=None):
        """
        врапер для функции python, вызываемой из плагина
        :param name: имя функции для вызова RPC
        :param module_path: путь к модулю, содержащему функцию
        :param help: документация. если не задана, то будет взята из комментария к функции
        :param fnc_name: имя функции ыв модуле. если не задано, будет использовано name
        :param module_name: имя модуля в списке модулей (чтобы не грузить модуль повторно) если не задано - будет взято из пути
        """
        module_path = Path(module_path)
        if not fnc_name:
            fnc_name = name
        if not module_name:
            module_name = module_path.name
            if module_name.endswith(".py"):
                module_name = module_name[:-3]
        m = PluginFnc.modules.get(module_name)
        if not m:
            m = SourceFileLoader(module_name, str(module_path)).load_module()
            PluginFnc.modules[module_name] = m
        if not hasattr(m, fnc_name):
            raise RuntimeError(f"Funcion '{fnc_name}' is not found in module '{module_name}'")
        f = getattr(m, fnc_name)
        super().__init__(name, f, help)


class SqlFnc(Fnc):

    def __init__(self, name, query, help=None):
        """
        врапер для sql-запроса, вызываемого как функция
        :param name: имя функции для вызова RPC
        :param query: запрос
        :param help: документация
        """
        if callable(query):
            if not help:
                help = getdoc(query)
            query = query()
        if isinstance(query, str):
            self.__query = query
            self.__args = ()
        elif isinstance(query, tuple):
            self.__query = query[0]
            self.__args = query[1:]
        super().__init__(name, help)

    async def __call__(self, env, **kwargs):
        ret = await env.sql(self.__query, *self.__args, **kwargs)
        return ret


import sys, os
sys.path.append(os.path.dirname(__file__))
from driver import Driver as BaseDriver, RowFactory
from psycopg import AsyncConnection
import psycopg.rows as rows


_FACTORY_ = {
    RowFactory.ANY.value: rows.tuple_row,
    RowFactory.TUPLE.value: rows.tuple_row,
    RowFactory.DICT.value: rows.dict_row,
    RowFactory.NAMED_TUPLE.value: rows.namedtuple_row,
}

def _col_type(col_type, type_cat):
    if type_cat == "S":
        return str
    elif type_cat == "N":
        if col_type in ("smallint", "integer", "bigint", "smallserial", "serial", "bigserial"):
            return int
        else:
            return float
    return str

class Driver(BaseDriver):

    LIKE = "ILIKE"

    def __init__(self, connection_string, on_open_close=None):
        """
        localhost:5432/postgres?user=postgres&password=postgres
        """
        super().__init__(connection_string, on_open_close)
        host, options = self.connection_string.split("/", 1)
        if ":" in host:
            host, port = host.split(":", 1)
            port = int(port.strip())
        else:
            port = None
        options = options.split("?", 1)
        dbname = options[0]
        if len(options) == 2:
            options = options[1].replace("&", " ")
        else:
            options = ""
        self._conn_str = f"host={host} dbname={dbname}"
        if port is not None:
            self._conn_str += f" port={port}"
        if options != "":
            self._conn_str += " " + options

    @property
    def in_transaction(self):
        return self._conn is not None

    async def begin(self):
        params = await self._before_open()
        self._conn = await AsyncConnection.connect(self.connection_string)
        for x in params:
            if x == "TIMEZONE":
                await self._conn.execute(f"set session timezone '{params[x]}'")
            else:
                await self._conn.execute(f"SET {x} = '{params[x]}'")

    async def sql(self, query, one_row=False, row_factory=RowFactory.DICT):
        if self._conn is None:
            await self.begin()
        async with self._conn.cursor() as acur:

            acur.row_factory = _FACTORY_[row_factory.value]

            await acur.execute(query)
            descr = acur.description
            if not descr:
                return acur.rowcount
            if one_row:
                ret = await acur.fetchone()
            else:
                ret = await acur.fetchall()
                if not ret:
                    ret = []
        return ret

    async def commit(self):
        if not self._conn:
            raise RuntimeError("Transaction is not open")
        await self._conn.commit()
        await self.rollback()

    async def rollback(self):
        if not self._conn:
            return
        await self._conn.close()
        await super().rollback()

    @staticmethod
    def _specify_type(defs):
        if defs.column_default and str(defs.column_default).startswith("nextval("):
            if defs.data_type == "integer":
                return BaseDriver.TypeSpec(type_name="serial", python_type=int, autoincrement=True)
            elif defs.data_type == "bigint":
                return BaseDriver.TypeSpec(type_name="bigserial", python_type=int, autoincrement=True)
            elif defs.data_type == "smallint":
                return BaseDriver.TypeSpec(type_name="smallserial", python_type=int, autoincrement=True)
        return BaseDriver._specify_type(defs)

    @staticmethod
    def ident_name(name: str) -> str:
        """
            Эмулирует поведение PostgreSQL функции quote_ident().
            """
        if name is None:
            return "NULL"
        # Проверяем, является ли строка валидным идентификатором
        if (name[0].isalpha() or name[0] == '_') \
                and all(char.isalnum() or char == '_' for char in name) \
                and name == name.lower():
            return name
        # Удваиваем существующие двойные кавычки
        name = name.replace('"', '""')
        # Заключаем в двойные кавычки
        return f'"{name}"'


from yaml import safe_load
from .class_tools import find_subclass
from collections import defaultdict
from pathlib import Path
from os import sep as pathsep
from collections import namedtuple

_ERRORS = {}

def load_errors(file_name):
    if pathsep in file_name:
        f = Path(file_name)
    else:
        f = Path(__file__).parent / file_name
    with f.open("r") as stream:
        data = stream.readlines()
    data = ''.join(data)
    data = safe_load(data)
    data = {x["code"]:x for x in data}
    _ERRORS.update(data)

class Error(Exception):

    def __init__(self, code, http_code:int, message:str):
        self._code = code
        self._http_code = http_code
        super().__init__(message)

    @property
    def code(self):
        return self._code

    @property
    def http_code(self)->int:
        return self._http_code

ErrorData = namedtuple("Error", ["code", "http_code", "message", "class_name"])

class BadError(Error):

    def __init__(self, code, http_code:int, not_found, *args, **kwargs):
        args = list(args)
        not_found = f"!!! UNKNOWN ERROR CLASS {not_found} !!!"
        if len(args) > 0:
            args[0] = f"{args[0]}\n{not_found}"
            args = tuple(args)
        else:
            args = (not_found,)
        super().__init__(code, http_code,*args, **kwargs)

def error_msg(code, *args, **kwargs)->ErrorData:
    global _ERRORS
    if not _ERRORS:
        load_errors("error.yaml")
    if isinstance(code, int):
        h_code = code
    else:
        code = str(code).upper()
        h_code = 501
    info = _ERRORS.get(code, {
        "http": 501,
        "msg": f"Unknown error\n{str(kwargs)}".replace("{", "/")
    })
    h_code = info.get("http", h_code)
    if msg:= info.get("msg"):
        msg = msg.format_map(defaultdict(lambda: "", kwargs))
    return ErrorData(code=code, http_code=h_code, message=msg, class_name=info.get("class"))

def raise_error(code, *args, **kwargs):

    err = error_msg(code, *args, **kwargs)

    if err.class_name and err.class_name!="Error":
        cls = find_subclass(BaseException, err.class_name)
        if cls is None:
            raise BadError(code, err.http_code, err.message, err.class_name, *args, **kwargs)
    else:
        cls = Error
    if cls is None or find_subclass(Error, cls.__name__):
        raise cls(code, err.http_code, err.message)
    raise cls(err.message, *args)



from .option import Option, DB, TIMEOUT, PAGE, CALLBACK, ITERATOR
from .driver import RowFactory
from ..error import raise_error

CONNECTION:str = ""

def set_connection(connection_string:str):
    global CONNECTION
    CONNECTION = connection_string

async def sql(query:str, *args, **kwargs):
    process = []
    db = callback = tm = row_factory = one_row = page = itr = None
    for n, arg in enumerate(args):
        if Option.is_option(arg):
            if not process:
                process.append(n)
            if one_row is None and arg.one_row is not None:
                one_row = arg
            elif one_row and arg.one_row is not None and one_row.one_row!=arg.one_row:
                raise_error("NEQ_ROWS", src1=one_row.__class__.__name__, src2=arg.__class__.__name__)
            if row_factory is None or row_factory.row_factory == RowFactory.ANY:
                row_factory = arg
            elif row_factory and arg.row_factory != RowFactory.ANY:
                raise_error("DEF_STR_FABRIC", src = row_factory.__class__.__name__)
            if arg.can_process:
                process.append(arg)
                continue
            if arg == DB:
                arg = DB()
            if isinstance(arg, DB) and db is None:
                db = arg
            elif isinstance(arg, CALLBACK) and callback is None:
                callback = arg
            elif isinstance(arg, DB):
                raise_error("DB_ALRDY_DEF")
            elif isinstance(arg, PAGE) and page is None:
                page = arg
            elif isinstance(arg, PAGE):
                raise_error("PAGE_ALRDY_DEF")
            elif isinstance(arg, TIMEOUT) and tm is None:
                tm = arg
            elif isinstance(arg, TIMEOUT):
                raise_error("TIMEOUT_ALRDY_DEF")
            elif arg == ITERATOR:
                itr = ITERATOR()
            elif isinstance(arg, ITERATOR):
                itr = arg
        elif process:
            raise_error("PMT_BEF_OPT")
    if process:
        args = args[:process.pop(0)]
    if db is None:
        db = DB(CONNECTION)
        in_self = True
    else:
        in_self = False
    await db.garbage_collect(False)
    one_row = one_row.one_row if one_row else False
    row_factory = row_factory.row_factory if row_factory and row_factory.row_factory != RowFactory.ANY else RowFactory.DICT
    if callback:
        kwargs["__PARAM_CALLBACK_GETTER"] = callback
    query = await db.connection.parse_query(query, *args, **kwargs)
    if page:
        query = await page(db.connection, query)
    if callback is not None:
        query = await callback(query)
    if itr:
        if page:
            raise_error("PAGE_ITERATOR")
        return itr(query, db, row_factory, process)
    try:
        if tm is None:
            ret = await db.connection.sql(query, one_row=one_row, row_factory=row_factory)
        else:
            ret = await tm(db.connection.sql(query, one_row=one_row, row_factory=row_factory))
        for f in process:
            ret, row_factory = await f.process(ret, db.connection, row_factory)
            if row_factory is not None:
                break
    except Exception as e:
        try:
            if db.connection.in_transaction:
                await db.connection.rollback()
        except Exception as e2:
            ...
        raise e
    if in_self and db.connection.in_transaction:
        await db.connection.commit()
    return ret


from typing import Optional
from datetime import datetime
from json import dumps, loads
from json5 import loads as json5_load
from yaml import safe_load as yaml_load
from toml import loads as toml_load
from pathlib import Path
from os import environ as ENV
import logging as LOG
from enum import Enum
from .timer import TimerList
from ..class_tools import classproperty
from ..error import raise_error, error_msg
import asyncio


class LogLevel(Enum):
    CRITICAL = LOG.CRITICAL
    FATAL = LOG.FATAL
    ERROR = LOG.ERROR
    WARNING = LOG.WARNING
    WARN = LOG.WARN
    INFO = LOG.INFO
    DEBUG = LOG.DEBUG
    SQL = LOG.DEBUG - 5
    NOTSET = LOG.NOTSET
    OFF = -1


class LogRecord(LOG.LogRecord):
    ext_data = {}
    # доп параметры в логер
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if len(args) > 1:
            self.sys_level = {
                LOG.CRITICAL: 2,
                LOG.FATAL: 2,
                LOG.ERROR: 3,
                LOG.WARNING: 4,
                LOG.INFO: 6,
                LOG.DEBUG: 7,
            }.get(args[1], 7)
        else:
            self.sys_level = 6
        for x in LogRecord.ext_data:
            setattr(self, x, LogRecord.ext_data[x])

def typed_val(key, values: dict):
    v = values.get(key)
    if v is None:
        return None
    elif "(" in key:
        key, t = key.split("(", 1)
        if not t.endswith(")"):
            return v
        match t.upper():
            case 'INT)':
                return int(v), key
            case 'STR)':
                return str(v), key
            case 'FLOAT)':
                return float(t), key
            case _:
                return v
    return v, key


def from_env(key:str, key_map:dict):
    ret = {key_map[x]: ENV[x] for x in key_map if x in ENV}
    if key and "." in key:
        eval_key = key.split(".")
        eval_key = [f"['{x}']" for x in eval_key]
        eval_key = "x" + "".join(eval_key)
    else:
        eval_key = f"x['{key}']"
    node = eval(eval_key, {"x": Config._settings}) if key else Config._settings[key]
    node.update(ret)

__CONF__ = None


class Config:
    _settings = None

    def __new__(cls, *args, **kwargs):
        global __CONF__
        if not __CONF__:
            __CONF__ = super().__new__(cls)
        return __CONF__

    @classproperty
    def initialized(cls):
        return cls._settings is not None

    def __init__(self, path: Optional[str | list[str] | dict] = None, env_map: Optional[dict] = None, can_include: Optional[list]=None) -> None:
        if Config.initialized:
            return

        Config._settings, env = {}, None

        def load(fn):
            if "://" in fn:
                x = asyncio.run(self._config_from_db(fn))
                return x
            with Path(fn).open() as f:
                data = f.readlines()
            data = ''.join(data)
            match Path(fn).suffix.lower():
                case ".json" | ".json5":
                    return json5_load(data)
                case ".yaml" | ".yml":
                    return yaml_load(data)
                case ".toml":
                    return toml_load(data)
                case _:
                    raise_error("BAD_FORMAT", format_name=fn)

        self._log_level = LogLevel.INFO
        if path is None:
            raise_error("BAD_CONFIG_PATH")
        elif isinstance(path, dict):
            Config._settings = path
        elif isinstance(path, str):
            Config._settings = load(path)
        elif isinstance(path, list):
            Config._settings = load(path.pop(0))
            for fn in path:
                data = load(fn)
                for x in data:
                    item = data[x]
                    dest = Config._settings.get(x)
                    if isinstance(dest, dict):
                        dest.update(item)
                    else:
                        data[x] = item

        if not Config._settings:
            Config._settings = {}
        if env_map:
            for x in env_map:
                from_env(x, env_map[x])
        if can_include:
            for x in can_include:
                f = Config._settings.get(x)
                if isinstance(f, str):
                    Config._settings[x] = load(f)
        if "logging" in Config._settings:
            self._logger = ...
            if self.logger:
                self.log(error_msg("INFO_LOG_START").message)
            del Config._settings["logging"]
        else:
            self._logger = None

        if "timers" in Config._settings:
            self._timers = TimerList()
            f = Config._settings["timers"]
            for x in f:
                self._timers.add(x, interval=f[x])
            del Config._settings["timers"]
        else:
            self._timers = None
        if "defaults" not in Config._settings:
            Config._settings["defaults"] = {}
        x = Config._settings.get("database")
        if isinstance(x, dict):
            Config._settings["defaults"]["database"] = tuple(x.keys())[0]

    @staticmethod
    def _cast(value, to_type):
        if value is None:
            return None
        elif to_type == "I":
            return int(value)
        elif to_type == "F":
            return float(value)
        elif to_type == "B" and value:
            return True
        elif to_type == "B":
            return False
        else:
            return str(value)

    async def _config_from_db(self, connection: str) -> dict:
        """
        считывает настройки из БД
        в параметрах соединения можно задать (внутри фигурных скобок, одной строкой без переносов, маленькими буквами)
        query - запрос для получения настроек, в запросе должны быть указаны поля в следующем порядке:
          1 - поле ключа
          2 - поле значения
          3 - идентификатор записи
          4 - идентификатор родительской записи
        также возможен вариант, когда
          4 - тип значенгия
          5 - идентификатор родительской записи
          в этом случае конфигуратор понимает следующие типы:
            I - int
            F - float
            S - str
            B - bool
          преобразование типа происходит в функции Config._cast(value, to_type) т.е. наследники Config могут обрабатывать
          другие типы.

        если в запросе указано where, то это условие будет добавлено во все запросы
        при построении конфигурации.

        по умолчанию запрос:
        select k, v, id, parent_id from config

        соответственно самый простой вариант таблицы:
        create table config(
            id serial not null primary key,
            parent_id integer references config(id),
            k varchar(150) not null,
            v varchar,
            constraint config_level_uk unique (parent_id, k)
        )
        """
        query = "select k, v, id, parent_id from config"
        if connection.endswith("}"):
            connection, params = connection.split("{", 1)
            params = loads(f"{{{params}")
            if x := params.get("query"):
                del params["query"]
                query = x
            if params:
                connection += dumps(params)
        if " where" in query:
            query, ext_where = query.rsplit(" where", 1)
            ext_where = f"\nand {ext_where}"
        else:
            ext_where = ""
        query, env = query.rsplit(" from ", 1)
        query, parent_id = query.rsplit(",", 1)
        parent_id = parent_id.strip()
        query += f", {parent_id} from {env}"
        try:
            from .db_context import DB_ENV, TUPLE
        except Exception as e:
            from db_context import DB_ENV, TUPLE
        env = DB_ENV(connection)

        async def load_level(where: str) -> dict:
            nonlocal env, query, parent_id, ext_where
            lst = await env.sql(query + "\nwhere " + where + ext_where, TUPLE)
            ret = {}
            with_tp = lst and len(lst[0]) >= 5
            for x in lst:
                childs = await load_level(f"{parent_id}={x[2]}")
                if childs:
                    ret[x[0]] = childs
                elif with_tp:
                    ret[x[0]] = self._cast(x[1], x[3])
                else:
                    ret[x[0]] = x[1]
            return ret

        ret = await load_level(f"{parent_id} is null")
        await env.commit()
        return ret

    def __getattr__(self, item):
        if not(item in Config._settings):
            raise_error("ATTR_NOT_FOUND", name=item)
        return Config._settings[item]

    @property
    def timers(self) -> TimerList:
        """
        Возвращает таймеры, перечисленные в секции timers
        таймер должен быть проинициализирован:
          init(callback, interval_type = TimerInterval.SECOND, interval: int = 0, immediately_start=True)
            calback       - обработчик таймера, может быть как async, так и синхронным
            interval_type - тип интервала (единица времени) MILLISECOND, SECOND, MINUTE, HOUR, DAY по умолчанию SECOND
            interval      - количество единиц времени, если задан, то перекроет считанное из конфига
            immediately_start - означает, что calback будет запущен сразу после инициализации, иначе через интервал
        например:
          cfg.init(callback, interval_type = TimerInterval.MINUTE):
        callbac должен быть функцией вида:
          def callback(name, prev_time):
              ...
          или:
          async def callback(name, prev_time):
              ...
          в параметр name передается в имя таймера, в prev_time предыдущее время запуска, первый раз prev_time = None
        дополнительные методы:
          start() - запускает таймер, если он остановлен или запущен с immediately_start=False
          stop()  - останавливает таймер. остановленный таймер можно зпустить методом start()
        :return: ab_engine.env.timer.Timer
        """
        return self._timers

    @staticmethod
    def hasattr(item):
        return item in Config._settings

    @staticmethod
    def _parse_log_level(level)->LogLevel:
        if isinstance(level, str):
            level = LogLevel[level.upper()]
        elif isinstance(level, int):
            level = LogLevel(level)
        elif isinstance(level, LogLevel):
            ...
        else:
            ValueError("Не понятен уровень лога")
        if level == LogLevel.SQL:
            level = LogLevel.DEBUG
        return level

    @property
    def log_level(self) -> LogLevel:
        return self._log_level

    @property
    def logger(self) -> LOG.Logger:
        if self._logger is not ...:
            return self._logger
        if not "level" in self.logging:
            self.logging["level"] = "info"
        if self.logging["level"].upper() == "SQL":
            self._log_level = LogLevel.SQL
        else:
            self._log_level = self._parse_log_level(self.logging["level"])
        self._logger = None
        if self._logger == LogLevel.OFF:
            return None
        LogRecord.ext_data = self.logging.get("ext_data", {})
        LOG.setLogRecordFactory(LogRecord)

        defs = {
            # '<%(sys_level)s>[%(service)s][%(asctime)s][%(levelname)s][%(filename)s:%(lineno)s]: %(message)s'),
            "format": self.logging.get("format", '[%(asctime)s][%(levelname)s][%(filename)s:%(lineno)s]: %(message)s'),
            "datefmt": self.logging.get("datefmt", '%m.%d.%Y %H:%M:%S'),
            "level": LOG.DEBUG if self._log_level.value == 5 else self._log_level.value
        }
        if not "handler" in self.logging:
            # Handler не задан ----------------------------------------------------------------------------------------
            for x in ("filename", "filemode"):
                v = self.logging.get(x)
                if v:
                    if x == "filename":
                        Path(v).parents[0].mkdir(parents=True, exist_ok=True)
                    defs[x] = v

            LOG.basicConfig(**defs)
            self._logger = LOG.getLogger()  # =logging.root
            return self._logger

        # Handler задан -----------------------------------------------------------------------------------------------
        lst = []

        def get_childs(cla):
            nonlocal lst
            lst.append(cla)
            for x in cla.__subclasses__():
                get_childs(x)

        hndl = self.logging["handler"]["class"]
        for x in LOG.Handler.__subclasses__():
            if x.__name__ == self.logging["handler"]:
                hndl = x
                break

        if isinstance(hndl, str):
            import logging.handlers

            get_childs(logging.Handler)

            for x in lst:
                if x.__name__ == hndl:
                    hndl = x
                    break
            if isinstance(hndl, str):
                raise_error("BAD_LOG_HANDLER", handler=self.logging['handler'])

        self._logger = logging.getLogger(self._name)

        params = {}
        for x in self.logging["handler"]:
            if x in ("class", "property"):
                continue
            v, x = typed_val(x, self.logging["handler"])

            if v is not None:
                if x == "filename":
                    Path(v).parents[0].mkdir(parents=True, exist_ok=True)
                params[x] = v
        def get_log_filename(filename):
            # Получаем директорию, где расположены логи
            log_directory = Path(filename).parent
            x = datetime.now().strftime("%Y%m%d")
            # Сформировали имя нового лог-файла
            filename = f"{log_directory / x}"
            x = f"{filename}.log"
            if not Path(x).exists():
                return x
            # Найдём минимальный индекс файла на текущий момент.
            index = 0
            f = f"{filename}.{index}.log"
            while Path(f).exists():
                index += 1
                f = f"{filename}.{index}.log"
            return f

        if self.logging["handler"]["class"] == "TimedRotatingFileHandler":
            x = params.get('filename')
            if x is None:
                raise_error("NA_FILE_NAME")
            Path(x).parents[0].mkdir(parents=True, exist_ok=True)
            with open(x, 'a'):
                ...
            hndl = hndl(**params)
            hndl.namer = get_log_filename
        else:
            hndl = hndl(**params)

        if "property" in self.logging["handler"]:
            for x in self.logging["handler"]["property"]:
                v, x = typed_val(x, self.logging["handler"]["property"])

                if v is not None:
                    if x == "filename":
                        Path(v).parents[0].mkdir(parents=True, exist_ok=True)
                    setattr(hndl, x, v)

        formatter = LOG.Formatter(
            defs["format"],
            defs["datefmt"]
        )
        hndl.setFormatter(formatter)

        self._logger.addHandler(hndl)
        return self._logger

    def log(self, msg, level:LogLevel | str | int=LogLevel.INFO, *args, **kwargs):
        """
        запись в лог
        :param msg: сообщение об ошибке. может быть экземпляром Exception
        :param level: уровень сообщения, для msg = Exception автомаптически вставляется ERROR. Может быть передан как строка
        :param args:  --/- зарезервировано для дальнейшего использования
        :param kwargs:-/
        """
        if self._log_level == LogLevel.OFF:
            return
        try:
            log_to = self.logger
        except Exception as e:
            log_to = None

        if isinstance(msg, Exception):
            level = LogLevel.ERROR
        else:
            level = self._parse_log_level(level)
            msg = str(msg)
        if level.value < self.log_level.value:
            return

        if not log_to:
            print(level.name, msg, flush=True)
            return

        try:
            params = {
                "name": log_to.name,
                "level": level.value,
                "fn": "",
                "lno": 0,
                "args": args,
                "exc_info": None,
                "func": "",
                "extra": None,
                "sinfo": None
            }
            if self._log_level.value <= LogLevel.DEBUG.value or isinstance(msg, Exception):
                if 'stacklevel' not in kwargs:
                    lvl = 2
                    info = log_to.findCaller(stacklevel=lvl)
                    p = str(Path(__file__).parent.parent)
                    while info[0].startswith(p):
                         lvl += 1
                         x = log_to.findCaller(stacklevel=lvl)
                         if '/venv/' in x[0]:
                             break
                         elif not x[0].startswith(p):
                             info = x
                             break
                         else:
                             info = x
                else:
                    info = log_to.findCaller(stacklevel=kwargs.get('stacklevel', 2))

                if isinstance(msg, Exception):
                    params["fn"] = info[0]
                else:
                    x = info[0].rsplit("/", 2)
                    del x[0]
                    params["fn"] = '/'.join(x)
                params["func"] = info[2]
                params["lno"] = info[1]

            if "timer" in kwargs:
                n = kwargs["timer"]
                x = self.__log_timers.get(n)
                if x:
                    x = datetime.now() - x
                    msg = f'{msg} [ {x} ]'
                self.__log_timers[n] = datetime.now()

            params["msg"] = f"{msg}"
            info = log_to.makeRecord(**params)
            log_to.handle(info)
        except Exception as e:
            # Больше вывести некуда - выводим на стандартный вывод
            print(level, msg, flush=True)
            print("LOGGER ERROR: ", str(e), flush=True)

    def db_connection(self, connection:str="")->str:
        if not Config.hasattr("database"):
            raise_error("NA_DB_IN_CONFIG")
        db = self.database
        if isinstance(db, str) and connection=="":
            return db
        if isinstance(db, dict):
            dflt = self.defaults if self.hasattr("defaults") else {}
            db = db.get(connection, db.get(dflt.get("database", "main")))
        if not db:
            raise_error("NA_DB", connection=connection)
        return db


import sys, os
sys.path.append(os.path.dirname(__file__))
from driver import Driver as BaseDriver, RowFactory
try:
    from mysql.connector.aio import connect
except ImportError:
    raise Exception("For use MySQL driver, you need to install mysql.connector\n$ pip install mysql-connector-python")
from collections import namedtuple

def as_is(description, data):
    return data

def as_dict(description, data):
    fields = [column[0] for column in description]
    if isinstance(data, list):
        for n, row in enumerate(data):
            data[n] = {key: value for key, value in zip(fields, row)}
        return data
    return {key: value for key, value in zip(fields, data)}

def as_namedtuple(description, data):
    fields = [column[0].lower() for column in description]
    cls = namedtuple("Row", fields)
    if isinstance(data, list):
        for n, row in enumerate(data):
            data[n] = cls._make(row)
        return data
    return cls._make(data)

_FACTORY_ = {
    RowFactory.ANY.value: as_is,
    RowFactory.TUPLE.value: as_is,
    RowFactory.DICT.value: as_dict,
    RowFactory.NAMED_TUPLE.value: as_namedtuple,
}


class Driver(BaseDriver):

    def __init__(self, connection_string, on_open_close=None):
        """
        localhost:3306/mysql?user=root&password=root
        """
        if "{" in connection_string:
            connection_string, options = connection_string.split("{", 1)

        super().__init__(connection_string, on_open_close)

        if "/" in self.connection_string:
            x, options = self.connection_string.split("/", 1)
        else:
            x, options = self.connection_string.split("?", 1)
        if ":" in x:
            x, y = x.split(":")
            self._conn_params = {
                "host": x,
                "port": int(y),
            }
        else:
            self._conn_params = {
                "host": x,
                "port": 3306,
            }
        if "?" in options:
            x, options = options.split("?",1)
            self._conn_params["database"] = x
        for x in options.split("&"):
            k, v = x.split("=", 1)
            self._conn_params[k] = v

    async def begin(self):
        await self._before_open()
        self._conn = await connect(**self._conn_params)

    async def sql(self, query, one_row=False, row_factory=RowFactory.DICT):
        if self._conn is None:
            await self.begin()
        cur = await self._conn.cursor()
        try:
            await cur.execute(query)
            descr = cur.description
            if descr:
                descr = descr.copy()
                if one_row:
                    data = await cur.fetchone()
                else:
                    data = await cur.fetchall()
            else:
                return cur.rowcount
        finally:
            try:
                await cur.close()
            except Exception as e:
                pass
        return _FACTORY_[row_factory.value](descr, data)

    async def commit(self):
        if not self._conn:
            raise RuntimeError("Transaction is not open")
        await self._conn.commit()
        await self.rollback()

    async def rollback(self):
        if not self._conn:
            return
        if x := self._on_open_close:
            await x(True)
        await self._conn.close()
        self._conn = None




from enum import Enum
from json import dumps
from datetime import datetime
from abc import ABC, abstractmethod
from collections import namedtuple

_is_option = None

def _set_is_option(f):
    global _is_option
    _is_option = f


class RowFactory(Enum):
    """
    Варианты фабрик для возврата набора данных
    """
    ANY = 0
    TUPLE = 1
    DICT = 2
    NAMED_TUPLE = 3


def reencode(s):
    if s is None:
        return " NULL"
    elif isinstance(s, str):
        return f"'{s}'"
    return s


class Driver(ABC):

    LIKE = "LIKE"

    def __init__(self, connection_string, on_open_close=None):
        self._conn = None
        if isinstance(connection_string, Driver):
            self._conn_str = connection_string.connection_string
            self._on_open_close = on_open_close or connection_string._on_open_close
            return
        self._on_open_close = on_open_close
        self._conn_str = connection_string

    @property
    def connection_string(self):
        return self._conn_str

    @property
    def in_transaction(self)->bool:
        """возвращает открыта ли транзакция"""
        return self._conn is not None

    async def _before_open(self)->dict:
        if self._conn:
            raise RuntimeError("Transaction already open")
        if x := self._on_open_close:
            params = await x(False)
            if not params:
                params = {}
        else:
            params = {}
        return params

    @abstractmethod
    async def begin(self):
        """открывает транзакцию"""
        ...

    @abstractmethod
    async def sql(self, query, one_row=False, row_factory=RowFactory.DICT):
        """выполняет запрос
        query текст запроса
        one_row если True, то выполняется fetchone иначе fetch
        row_factory способ представления результата
        если набор данных пустой и one_row==False - возвращает []
        если набор данных пустой и one_row==True - возвращает None
        если запрос не возвращает результат - возвращает число обработанных строк (целое)
        """
        ...

    @abstractmethod
    async def commit(self):
        """подтверждает транзакцию и закрывает соединение"""
        ...

    async def rollback(self):
        """откатывает транзакцию и закрывает соединение"""
        if x:=self._on_open_close:
            await x(True)
        self._conn = None

    @staticmethod
    def var_to_sql(var, str_after=" "):
        if var is None:
            return " NULL"
        elif isinstance(var, str):
            var = var.replace("'", "''")
            return reencode(var)
        elif isinstance(var, (list, tuple)):
            if str_after[:6].lower() == '::json' or str_after[0] != ':':
                var = dumps(var, ensure_ascii=False).replace("'", "''")
                return reencode(var)
            else:
                var = ','.join(['NULL' if x is None else str(x) for x in var])
                return reencode(var)
        elif isinstance(var, set):
            var = ','.join('NULL' if x is None else str(x) for x in var)
            return reencode(var)
        elif isinstance(var, dict):
            var = dumps(var, ensure_ascii=False).replace("'", "''")
            return reencode(var)
        elif isinstance(var, datetime):
            return f" '{var}'"
        else:
            return f" {var}"

    async def parse_query(self, query, *args, **kwargs):
        """подставляет параметры в текст запроса"""
        if callback:=kwargs.get("__PARAM_CALLBACK_GETTER"):
            del kwargs["__PARAM_CALLBACK_GETTER"]
        qs = None
        for x in query.split("$"):
            if qs is None:
                qs = x
                continue
            n, p = '', 0
            if str(x[p]).isdigit():
                while p < len(x) and x[p].isdigit():
                    n += x[p]
                    p += 1
                n = int(n) - 1
                if n < 0 or n >= len(args):
                    raise AttributeError(f"Attribute with index {n} is not exists")
                n = args[n]
            else:
                while p < len(x) and not(x[p].isspace() or x[p] in "(:,)"):
                    n += x[p]
                    p += 1
                if n in kwargs:
                    n = kwargs[n]
                elif callback:
                    n = callback[n]
                else:
                    raise AttributeError(f"Attribute with name {n} is not exists")
            qs += f"{self.var_to_sql(n, x)}{x[p:]}"
            
        if qs:
            query = qs

        return query

    @staticmethod
    def is_option(what):
        global _is_option
        if _is_option:
            return _is_option(what)
        else:
            return None

    async def parse_func(self, func_name, *args, **kwargs):
        """
        Возвращает запрос для вызова функции func_name
        :param args: неименованные параметры
        :param kwargs: именованные параметры
        :return: запрос
        """
        ret = f"select {func_name}("
        for n in range(len(args)):
            if self.is_option(args[n]):
                break
            if n > 0:
                ret += ", "
            ret += f"${n+1}"
        for n, x in enumerate(kwargs):
            if n > 0:
                ret += ", "
            ret += f"${x}"
        ret += ")"
        return await self.parse_query(ret, *args, **kwargs)

    async def cast(self, param_name, to_type):
        """
        возвращает часть запроса для приведения типа
        :param param_name:
        :param to_type:
        :return:
        """
        return f"{param_name}::{to_type}"

    @staticmethod
    def ident_name(name:str)->str:
        return name

    TypeSpec = namedtuple("TypeSpec", ["type_name", "python_type", "autoincrement"])

    @staticmethod
    def _specify_type(defs):
        if defs.data_type in ("smallint", "integer", "bigint", "smallserial", "serial", "bigserial", "int"):
            pt = int
        elif defs.data_type in ("double precision", "numeric", "real", "decimal", "money", "float"):
            pt = float
        else:
            pt = str
        return Driver.TypeSpec(type_name=defs.data_type, python_type=pt, autoincrement=False)

    async def table_struct(self, table_name) -> dict:
        """
        возвращает структуру заданной таблицы
        """
        if not self.in_transaction:
            await self.begin()
        schema_name, table_name = f"{table_name}.".split(".", 1)
        if table_name=="":
            table_name = self.ident_name(schema_name)
            schema_name = None
        else:
            table_name = self.ident_name(table_name[:-1])
            schema_name = self.ident_name(schema_name)
        fltr = f"\nwhere ku.table_name='{table_name}'"
        if schema_name is not None:
            fltr += f" and ku.table_schema='{schema_name}'"
        qry =f"""select ku.column_name, ku.data_type, ku.is_nullable, ku.character_maximum_length, 
                       ku.numeric_precision, ku.numeric_scale, ku.column_default
                from information_schema.columns ku {fltr}
                order by ku.ordinal_position"""
        qry = await self.sql(qry, one_row=False, row_factory=RowFactory.NAMED_TUPLE)
        if not qry:
            return {}
        fields = []
        for x in qry:
            t = self._specify_type(x)
            field = {
                "name": x.column_name,
                "type": t.type_name,
                "not_null": x.is_nullable is not None and f'{x.is_nullable} '[0].upper() in "FNН",
                "python_type": t.python_type,
            }
            if t.autoincrement:
                field["autoincrement"] =True
            n = self.ident_name(field["name"])
            if n!=field["name"]:
                field["field"] = n
            if x.character_maximum_length:
                field["size"] = x.character_maximum_length
            elif field["python_type"]!=int and x.numeric_precision:
                n = f'{x.numeric_precision}.{x.numeric_scale}'
                field["size"] = float(n)
            fields.append(field)
        qry = f"""select
            x.constraint_name,
            x.constraint_type,
            x.column_name,
            x.reference_schema,
            x.reference_table,
            r.column_name reference_column
        from (
            select ku.constraint_name, tc.constraint_type, ku.column_name,
                   rt.table_schema reference_schema, rt.table_name reference_table,
                   row_number() over (partition by ku.constraint_name order by ku.ordinal_position) column_position
            from information_schema.key_column_usage ku
            inner join information_schema.table_constraints tc on tc.constraint_name = ku.constraint_name and tc.table_name = ku.table_name  and tc.table_schema = ku.table_schema
            left join information_schema.referential_constraints rc on rc.constraint_name = ku.constraint_name
            left join information_schema.table_constraints rt on rc.unique_constraint_name = rt.constraint_name
            {fltr}
            order by ku.constraint_name, ku.ordinal_position
            ) x
        left join (
            SELECT
                    tc.table_name, tc.table_schema, i2.COLUMN_NAME,
                    row_number() over (partition by tc.table_name, tc.table_schema order by  i2.ordinal_position) column_position
                FROM
                    information_schema.table_constraints tc
                JOIN
                    information_schema.key_column_usage i2
                    ON tc.constraint_name = i2.constraint_name
                WHERE
                    tc.constraint_type = 'PRIMARY KEY'
        ) r on x.reference_schema = r.table_schema and x.reference_table = r.table_name and x.column_position=r.column_position
        order by x.constraint_name, x.column_position
        """
        qry = await self.sql(qry, one_row=False, row_factory=RowFactory.NAMED_TUPLE)
        constraints = []
        item = {"name": None}
        for x in qry:
            if x.constraint_name!=item["name"]:
                if item["name"] is not None:
                    constraints.append(item)
                t = x.constraint_type[0].upper()
                if t == "P":
                    item = {
                        "name": x.constraint_name,
                        "fields": [],
                        "type": "primary key"
                    }
                elif t == "U":
                    item = {
                        "name": x.constraint_name,
                        "fields": [],
                        "type": "unique"
                    }
                elif t == "F":
                    item = {
                        "name": x.constraint_name,
                        "fields": {},
                        "table": f"{self.ident_name(x.reference_schema)}.{self.ident_name(x.reference_table)}",
                        "type": "foreign key"
                    }
                else:
                    item = {"name": None}
                    continue
            if item["type"][0] == "f":
                item["fields"][x.column_name] = x.reference_column
            else:
                item["fields"].append(x.column_name)
        if item["name"]:
            constraints.append(item)
        return {
            "table": f"{schema_name}.{table_name}" if schema_name else table_name,
            "fields": fields,
            "constraints": constraints
        }

    async def page(self, query, limit, offset):
        query = f"{query}\nlimit {limit}"
        if offset:
            query = f"{query}\noffset {offset}"
        return query

    async def get_position(self, table, key, filter):
        """
        возвращает номер первой строки table, отсортированной по key, соответствующей фильтру filter
        """
        n_field = "co_lu_mn__po_si_ti_on__by__ke_y"
        qry = f"""select row_number() over (order by {key}) {n_field}, t.* from {table} t"""
        qry = f"""select {n_field} from ({qry}) x where {filter} order by {n_field}"""
        x = await self.sql(qry, one_row=True, row_factory=RowFactory.TUPLE)
        if x:
            return x[0]



from typing import Callable, Optional
from ..db import DB, sql, JSON, CALLBACK, TUPLE
from .config import Config, LogLevel
from json5 import loads, dumps
from ..db.table import EnvTable
from inspect import iscoroutinefunction
from ..error import raise_error, Error


class Property:

    def __init__(self, getter: Callable, setter: Callable ):
        self._getter = getter
        self._setter = setter

    @property
    def value(self):
        return self._getter()

    @value.setter
    def value(self, value):
        self._setter(value)


class DB_ENV:

    def __init__(self, connection: str="", db_params:Optional[set]=None, **kwargs):
        """
        Окружение для работы с БД
        :param connection: строка соеддинения или экземпляр DB_ENV, на основе которого нужно создать данный
        :param db_params: список переменных, которые должны передаваться в в окружение соединения с БД
        :param kwargs: значения переменных
        """
        self._db_params = set()
        self._params = {}
        if isinstance(connection, DB_ENV):
            env = connection
            self._db_params = env._db_params.copy()
            self._params = env._params.copy()
            connection = env.connection_string
        elif isinstance(connection, DB):
            connection = DB.connection.connection_string
        if db_params:
            self._db_params.update(db_params)
        self._params.update(kwargs)

        if "://" not in connection:
            connection = Config().db_connection(connection)
        if "{" in connection and connection.endswith("}"):
            connection, x = connection.split("{", 1)
            p = loads(f"{{{x}")
            for x in p:
                if not x in self._db_params:
                    self._db_params.add(x)
            if self._params:
                p.update(self._params)
            self._params = p
        self._connection_str = connection
        self._context = None
        self._on_commit = set()

    @property
    def connection_string(self):
        # возвращает строку соединения
        return self._connection_str

    def on_commit(self, callback:Callable, subscribe=True):
        """
        Позволяет добавить или удалить обработчик события подтверждения транзакции
        :param callback: функция, в которую первым атрибутом передается cсылка на окружение
                        если функция вернет False, то commit будет отменен и транзакция откатится (rollback)
        :param subscribe: True указывает, что данный обработчик подписывается на событие (по умолчанию)
        """
        if subscribe:
            self._on_commit.add(callback)
        elif callback in self._on_commit:
            self._on_commit.remove(callback)

    def __getitem__(self, key):
        if not self.has_item(key):
            raise_error("ENV_ATTR_NOT_EXISTS", name=key)
        ret = self._params.get(key)
        return ret.value if isinstance(ret, Property) else ret

    def __setitem__(self, key, value):
        x = self._params.get(key)
        if isinstance(x, Property):
            x.value = value
        else:
            self._params[key] = value

    def has_item(self, key):
        # возвращает True если в данном DB_ENV хранится переменная с именем key
        return key in self._params

    async def __aenter__(self):
        if self._context is not None:
            raise_error("ENV_CONTEXT_EXISTS")
        if self._db_params:
            s = self._connection_str + f"{{{dumps({x:self[x] for x in self._db_params if self.has_item(x)})}}}"
        else:
            s = self._connection_str
        self._context = DB(s)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self._context:
            return
        try:
            if not self._context.connection.in_transaction:
                return
            if exc_tb:
                await self.rollback()
            else:
                await  self.commit()
        finally:
            self._context = None

    async def sql(self, query, *args, **kwargs):
        # выполняет команду sql. если на момент вызова соединение с БД (транзакция) закрыто, то создает новое соединение

        def log_callback(q_text):
            cfg=Config()
            if cfg.log_level.value <= LogLevel.SQL.value and cfg.log_level != LogLevel.OFF:
                cfg.log(q_text, LogLevel.SQL)

        if self._context is None:
            await self.__aenter__()

        cmd, query = f"{query} ".split(" ", 1)
        cmd, query = cmd.strip().lower(), query.strip()
        if cmd=="*":
            query = "select * from\n" + query
        elif cmd==r"\call":
            query = await self._context.connection.parse_func(query, **kwargs)
        elif cmd==r"\d":
            return await self._context.connection.table_struct(query.strip())
        elif cmd=="\json":
            query, params = query.split("(", 1)
            if params.lower() in ("json)", "jsonb)"):
                args = (kwargs,) + args
                kwargs = {}
                params = await self._context.connection.cast("$1", params[:-1].lower())
            args = args + (JSON,)
            key = "P_RA+MS_" + str(id(self))
            query = await self._context.connection.parse_func(query, key)
            query = query.replace(f"'{key}'", params)
        elif cmd:
            query = f"{cmd} {query}"

        args = args + (self._context, CALLBACK(callback_function=log_callback, attribute_collection=self),)
        ret = await sql(query,*args, **kwargs)
        return ret

    @property
    def in_transaction(self):
        # возвращает True если в данном DB_ENV открыта транзакция
        return self._context.connection.in_transaction if self._context is not None else False

    async def commit(self):
        # подтверждает транзакцию и закрывает соединение с БД
        try:
            for x in list(self._on_commit):
                if iscoroutinefunction(x):
                    x = await x(self)
                else:
                    x = x(self)
                if x == False and self.in_transaction:
                    await self.rollback()
                    break
        except Exception as e:
            if self.in_transaction:
                self.rollback()
            raise e

        if self.in_transaction:
            await self._context.connection.commit()
        self._context = None

    async def rollback(self):
        # откатывает транзакцию и закрывает соединение с БД
        if self.in_transaction:
            await self._context.connection.rollback()
        self._context = None

    async def table(self, name, page_size=100, async_delay=0.000001, if_not_exists:Optional[list]=None):
        """
        :param name: имя таблицы
        :param page_size: количество строк таблицы, одновременно находящихся в памяти
        :param async_delay: задержка итерации цикла при использовании курсора как итератора
        :param if_not_exists: список команд, которые вызываются, если тавблица не существует
        :return: курсор для работы с таблицей
        """
        if self._context is None:
            await self.__aenter__()
        try:
            t = await EnvTable.create(name, self, page_size=page_size, async_delay=async_delay)
            return t
        except Error as e:
            if not e.code=="NA_TABLE" or not if_not_exists:
                raise e
        if isinstance(if_not_exists, str):
            if_not_exists = [if_not_exists]
        for x in if_not_exists:
            await self.sql(x)
        t = await EnvTable.create(name, self, page_size=page_size, async_delay=async_delay)
        return t

    @staticmethod
    async def garbage_collect():
        """
        сборка мусора и закрытие подвисших соединений
        """
        DB.garbage_collect()


from codecs import namereplace_errors

from .rpc import Fnc, call_rpc
from ..error import raise_error
from ..env import DB_ENV, Config, LogLevel
from json5 import loads as json_loads, dumps as json_dumps
from asyncio import gather
from ..error import error_msg


class JSON_RPC:

    @staticmethod
    def rpc_error(code, data=None, header=None):

        err = error_msg(code)
        if err.message is None:
            err.message = "Server error (UNKNOWN!!)"
            if not isinstance(code, int) or code > -32000 or code < -32099:
                code = -32000
            err.code = code
        code = {
            "code": err.code,
            "message": err.message,
        }
        if data:
            code["data"] = data if isinstance(data, (dict, list)) else str(data)
        code = {"error": code}
        if header:
            if "id" in header:
                code["id"] = header["id"]
        else:
            header = {}
        code["jsonrpc"] = header.get("jsonrpc", "2.0")
        return code

    def __init__(self, connection = None):
        self._con = connection

    async def _rpc(self, env, method, header, params):
        # можно перезагрузить в наследниках, чтобы например, вызывать команды мз другого списка
        f = Fnc.search(method)
        if f is None:
            return self.rpc_error(-32601, header=header)
        if isinstance(params, list):
            params = {"LIST_OF_PARAMS": params}
        elif not isinstance(params, dict):
            raise_error("BAD_RPC_FORMAT", name=method)
        return await call_rpc(f, env, **params)

    async def _call_one(self, env, message):
        if "method" not in message:
            return self.rpc_error(-32600, header=message)
        if "params" in message:
            params = message["params"]
            del message["params"]
        else:
            params = {}
        f = message['method']
        self.log(f"CALL:{f}")
        try:
            res = await self._rpc(env, f, message, params)
            if isinstance(res, tuple):
                res = list(res)
            elif not isinstance(res, (dict, list, str)):
                res = str(res)
            elif not (isinstance(res, dict) and len(res) == 1 and "error" in res):
                res = {"result": res}
            res["jsonrpc"]=message.get("jsonrpc", "2.0")
            if res["jsonrpc"]=="1.0":
                if "result" not in res:
                    res["result"] = None
                if "error" not in res:
                    res["error"] = None
            if "id" in message:
                res["id"] = message["id"]
        except Exception as e:
            res = self.rpc_error(-32000, data=e, header=message)
        if res.get("error") is not None:
            self.log(f"{f}: ERROR:\n{res['error']}!!")
            if env.in_transaction:
                await env.rollback()
        else:
            self.log(f"{f}: OK!")
        return res

    async def _call_part(self, message, ext_params=None):
        env = self._get_env(message=message, ext=ext_params,  only_new=True)
        ret = await self._call_one(env, message)
        if env.in_transaction:
            await env.commit()
        return ret

    def _get_env(self, message, ext, only_new=False):
        if not only_new and isinstance(self._con, DB_ENV):
            return self._con
        return DB_ENV(self._con, MESSAGE=message, EXT=ext)

    async def __call__(self, message, **kwargs):
        """
        Выполняет команду в соответствии с правилами JSON-RPC
        :param message: Тело сообщения, которое может быть представлено str, dict, list или tuple
                        Eсли сообщение является строкой, то сначала эта строка преобразуется в структуру python по
                        правилам json5. В результате преобразования сообщение станет dict или list
                        Если сообщение dict, то подразумевается, что в нем есть атрибут method, содержащий имя метода,
                        а также могут быть параметры:
                          params - параметры метода
                          id - идентификатор сообщения
                          прочие параметры, которые будут проигнорированы механизмом RPC
                        Eсли сообщение list, то каждый элемент списка должен быть dict, который будет обработан по правилам,
                        описанным выше. При этом, все dict переданные в списке будут обрабатываться параллельно. На
                        выходе будет получен list с результатами обработки каждого dict.
                        Если сообщение tuple, то оно будет выполнено почти как list, с той разницей, что dict будут
                        выполнены последовательно.
        :param kwargs:  Дополнительные параметры, которые будут доступны методам в параметре окружения EXT,
                        также в параметре MESSAGE доступно тело обрабатываемого сообщения (dict)
        :return:        список ответов, либо ответ (dict) содержащий атрибут result с ответом, либо error с сообщением
                        об ошибке.
        """
        if isinstance(message, str):
            try:
                message = json_loads(message)
            except Exception as e:
                return self.rpc_error(-32700, e)
            to_str = True
        else:
            to_str = False
        if isinstance(message, list):
            res = [self._call_part(x, ext_params=kwargs) for x in message]
            res = await gather(*res)
        elif isinstance(message, tuple):
            env = self._get_env(message=message, ext=kwargs)
            res = []
            for x in message:
                x = await self._call_one(env, x)
                res.append(x)
                if not env.in_transaction and isinstance(x, dict) and len(x) == 1 and "error" in x:
                    break
            if env.in_transaction:
                await env.commit()
        elif isinstance(self._con, DB_ENV):
            res = await self._call_one(self._con, message)
        else:
            env = self._get_env(message=message, ext=kwargs)
            res = await self._call_one(env, message)
            if env.in_transaction:
                await env.commit()
        return json_dumps(res, indent=4) if to_str else res

    @staticmethod
    def log(msg, level=LogLevel.DEBUG, *args, **kwargs):
        Config().log(msg, level, *args, **kwargs)


async def call_json(message:str|dict|list, connection="", ext_params=None):
    rpc = JSON_RPC(connection)
    ret = await rpc(message, ext_params=ext_params)
    return ret



from enum import Enum
from datetime import datetime, timedelta
from inspect import iscoroutinefunction
from ..error import raise_error
import asyncio

class TimerInterval(Enum):
    MILLISECOND = 1
    SECOND = 1000 * MILLISECOND
    MINUTE = 60 * SECOND
    HOUR = 60 * MINUTE
    DAY = 24 * HOUR

class Timer:

    def __init__(self, interval: int = 10, interval_type = TimerInterval.SECOND, name:str=""):
        self._last_tm = None
        self._task = None
        self._callback = None
        self._interval_type = interval_type
        self._interval = abs(interval)
        self._name = name

    async def _do_task(self):
        chk = self._interval_type.value * self._interval
        t = (chk / 4) // 1000
        chk = timedelta(milliseconds=chk)
        if self._last_tm:
            if iscoroutinefunction(self._callback):
                await self._callback(self.name, None)
            else:
                self._callback(self.name, None)
        else:
            self._last_tm = datetime.now()
        while self._task is not None:
            await asyncio.sleep(t)
            if (datetime.now() - self._last_tm) < chk:
                continue
            if self._task is None:
                break
            try:
                if iscoroutinefunction(self._callback):
                    await self._callback(self.name, self._last_tm)
                else:
                    self._callback(self.name, self._last_tm)
            except Exception as e:
                ...
            finally:
                self._last_tm = datetime.now()

    @property
    def name(self):
        return self._name

    def init(self, callback, interval_type = TimerInterval.SECOND, interval: int = 0, immediately_start=True):
        if isinstance(interval_type, str):
            interval_type = TimerInterval[interval_type.upper()]
        if self._task:
            self._task = None
        if interval:
            self._interval = interval
        if interval_type:
            self._interval_type = interval_type
        self._callback = callback
        if immediately_start:
            self.start()

    def start(self, immediately_start=True):
        if self._callback is None:
            raise_error("TMR_WO_CALLBACK", name = self._name)
        if self._task:
            raise_error("TMR_ARDY_STARTED", name = self._name)
        if immediately_start:
            self._last_tm = datetime.now()
        else:
            self._last_tm = None
        self._task = asyncio.create_task(self._do_task())

    def stop(self):
        if self._task is None:
            raise_error("TMR_ARDY_STOPPED", name = self._name)
        self._task = None

    @property
    def started(self):
        return self._task is not None


class TimerList:

    def __init__(self):
        self._timers = {}

    def add(self, name:str, interval: int = 10, interval_type = TimerInterval.SECOND, callback=None, immediately_start=True):
        if name in self._timers:
            raise_error("TMR_ARDY_EXISTS", name=name)
        self._timers[name] = Timer(interval=interval, interval_type=interval_type, name=name)
        if callback:
            self._timers[name].init(callback=callback, immediately_start=immediately_start)

    def __getitem__(self, item)->Timer:
        if not item in self._timers:
            raise_error("TMR_NOT_EXISTS", name=item)
        return self._timers[item]


from ..class_tools import ReadOnlyPropDict
from .option import DB, PAGE, ONE, ROW, DICT, TUPLE, OBJECT, ALL
from ..error import raise_error
from .processor import sql
import asyncio


class ROW(ReadOnlyPropDict):

    def __str__(self):
        s = ""
        for x in self._data:
            v = str(self._data[x].value).replace("\t","    ")
            s += f" {x}: {v}" + "\t"
        return s[:-1]

    def __getitem__(self, item):
        if item not in self._data:
            raise_error("FIELD_NOT_FOUND", name=item)
        return self._data[item]


class Field:

    class _Condition:

        class Name(str):
            ...

        @staticmethod
        def _lquot(p_type, v):
            if v is None:
                return "NULL"
            elif isinstance(v, Field._Condition.Name):
                return v
            v = str(v)
            if p_type not in (int, float, bool):
                v = f"""'{v.replace("'","''")}'"""
            return v

        def __init__(self, left, op, right):
            if isinstance(right, Field):
                if right._table == left._table:
                    right = Field._Condition.Name(right.name)
                right = right.value
            if not isinstance(left, Field):
                raise_error("NOT_LFIELD")
            elif op in ("%", "//"):
                x = left._table._db.connection.LIKE
                op = x if op == "%" else f"NOT {x}"
                if left._p_type!=str:
                    raise_error("BAD_OP_FOR", op=op, name=str(left._p_type))
                elif not isinstance(right, (str, Field._Condition.Name)):
                    raise_error("BAD_OP_FOR", op=op, name=str(right))
                right = self._lquot(left._p_type, right)
            elif isinstance(right, (list, set, tuple)):
                if op not in ("==", "!=", "in"):
                    raise_error("BAD_OP_FOR", op=op, name=str(right))
                elif op == "!=":
                    op = "not in"
                else:
                    op = "in"
                right = ",".join([self._lquot(left._p_type, x) for x in right])
            elif right is None:
                if op not in ("==", "!="):
                    raise_error("BAD_OP_FOR", op=op, name=str(right))
                elif op == "==":
                    op = "is"
                else:
                    op = "is not"
                self._qry = f"{left()} {op} NULL"
                return
            else:
                right = self._lquot(left._p_type, right)
            self._qry = f"{left()} {op} {right}"

        def __call__(self):
            return self._qry

    def __init__(self, defs: dict, table):
        self._table = table
        self._p_type = defs.get("python_type")
        self._db_type = defs.get("type")
        self._name = defs.get("name")
        self._not_null = defs.get("not_null", False)
        x = defs.get("field")
        if x and self._name!=x:
            self._in_db_name=defs.get("field")

    def __repr__(self):
        return f"{self.table.name}.{self.name}"

    @property
    def table(self):
        """ссылка на таблицу, к которой относится поле"""
        return self._table

    @property
    def name(self):
        """имя поля"""
        return self._name

    @property
    def python_type(self):
        """тип python, соответствующий значениям поля"""
        return self._p_type

    @property
    def not_null(self):
        """признак того, что поле не может принимать значение None"""
        return self._not_null

    @property
    def value(self):
        """возвращает значение поля"""
        ptr = self._table._ptr
        buf = self._table._page
        if buf is None or len(buf)<=ptr or ptr<0:
            return None
        return buf[ptr][self.name]

    @value.setter
    def value(self, value):
        """устанавливает значение поля"""
        if self.table._key == 1:
            raise_error("NA_PK", name = self._table.name)
        if self.python_type is None:
            raise_error("NOT_ERD_DT", name = self._db_type)
        ptr = self.table._ptr
        buf = self.table._page
        if buf is None or len(buf) <= ptr or ptr < 0:
            raise_error("BAD_CURSOR", table=self.table.name)
        if value is None and self._not_null:
            raise_error("NOT_NULL", name=self.name)
        else:
            value = self.python_type(value)
        if buf[ptr][self.name]!=value:
            self.table._field_changed(ptr, self.name)
            buf[ptr][self.name] = value

    def __lt__(self, other):
        return Field._Condition(self, "<", other)

    def __le__(self, other):
        return Field._Condition(self, "<=", other)

    def __eq__(self, other):
        return Field._Condition(self, "=", other)

    def __ne__(self, other):
        return Field._Condition(self, "!=", other)

    def __gt__(self, other):
        return Field._Condition(self, ">", other)

    def __ge__(self, other):
        return Field._Condition(self, ">=", other)

    def __contains__(self, other):
        return Field._Condition(self, "in", other)

    def __mod__(self, other):
        return Field._Condition(self, "%", other)

    def __floordiv__(self, other):
        return Field._Condition(self, "//", other)

    def __call__(self, *args, **kwargs):
        if hasattr(self, "_in_db_name"):
            return self._in_db_name
        return self.name

    async def _agg_query(self, field_agg):
        q = f"select {field_agg} from {self.table.name}"
        if self.table._filter:
            q = f"{q} where {self.table._filter}"
        ret = await sql(q, self.table._db, ONE)
        if self.table._auto_close_conn:
            self.table._db.rollback()
        return ret

    async def min(self):
        """возвращает минимальное значение поля с учетом текущего фильтра таблицы"""
        return await self._agg_query(f"min({self()})")

    async def max(self):
        """возвращает максимальное значение поля с учетом текущего фильтра таблицы"""
        return await self._agg_query(f"max({self()})")

    async def avg(self):
        """возвращает среднее значение поля с учетом текущего фильтра таблицы"""
        return await self._agg_query(f"avg({self()})")

    async def sum(self):
        """возвращает сумму значений поля с учетом текущего фильтра таблицы"""
        return await self._agg_query(f"sum({self()})")

    async def count(self):
        """возвращает количество уникальных значений поля с учетом текущего фильтра таблицы"""
        return await self._agg_query(f"count(distinct {self()})")


class Table:

    class _FILTER(ALL):
        ...

    def __init__(self, table_struct:dict, db:DB, page_size=100, auto_close_conn=True, async_delay=0.001):
        self._table = table_struct["table"]
        self._db = db
        self._delay = async_delay
        self._page_size = page_size
        self._auto_close_conn = auto_close_conn
        self._ptr = None
        self._changed = set()
        self._offset = 0
        self._page = None
        self._filter = ""
        self._row = ROW(**{x["name"]:Field(x, self) for x in table_struct.get("fields", [])})
        self._key = 1
        constraint = table_struct.get("constraints")
        for x in  constraint:
            if x["type"]=="primary key":
                self._key_fields = x["fields"]
                self._key = ', '.join(self._row[f]() for f in self._key_fields)
                break
        if self._key == 1:
            for x in constraint:
                if x["type"] == "unique":
                    self._key_fields = x["fields"]
                    self._key = ', '.join(self._row[f]() for f in self._key_fields)
                    break

    @property
    def key(self):
        """
        :return:  имя ключа или список полей составного ключа
        """
        if self._key==1:
            return None
        if isinstance(self._key_fields, list) and len(self._key_fields)==1:
            return self._key_fields[0]
        return self._key_fields

    @key.setter
    def key(self, value):
        """
        позволяет установить имя ключа или список имен полей составного ключа, отличный от
        полученного из БД при инициализации таблицы
        """
        if isinstance(value, Field):
            value = [value.name,]
        elif isinstance(value, str):
            value = [value,]
        elif isinstance(value, (list,tuple,set)):
            value = list(value)
        else:
            raise_error("ER_KEY_LST")
        for x in value:
            self.row[x]
        self._key_fields = value
        self._key = ', '.join(self._key_fields)

    def _field_changed(self, ptr, field):
        self._changed.add(ptr)

    @property
    def field_names(self)->tuple:
        """
        :return:  tuple с именами полей таблицы
        """
        return tuple(self.row._data.keys())

    @property
    def name(self):
        """
        :return: имя таблицы
        """
        return self._table

    @property
    def row(self) -> ROW:
        """
        поля текущей строки таблице, к полям можно оброщаться через строковое имя в квадратных скобках, либо по имени через точку
        т.е. t.row.id и t.row["id"] это одно и то же
        """
        return self._row

    @classmethod
    async def create(cls, table_name, db:DB, page_size=100, auto_close_conn=True, async_delay=0.0001):
        """
        Создает объект для работы с указанной таблицей.
        При работе таблица сортируется по ключу, если ключа нет, то по первому полю
        :param table_name: Иья таблицы
        :param db: опция DB для соединения с БД
        :param page_size: размер окна для чтения таблицы
        :param auto_close_conn: если True, то соединение автоматически закрывается после каждого чтения или записи таблицы
        :param async_delay: время передачи управления другим процессам при использовании таблицы как итератора
        :return:
        """
        table_struct = await db.connection.table_struct(table_name)
        if table_struct is None:
            raise_error("NA_TABLE", name=table_name)
        if auto_close_conn:
            await db.connection.rollback()
        t = cls(table_struct, db, page_size, auto_close_conn, async_delay)
        await t.first()
        return t

    @property
    def position(self):
        """строка, на которой стоит курсор (счет с 0)"""
        return self._offset + (self._ptr or 0)


    async def seek(self, *args, **kwargs):
        if kwargs or len(args) != 1 or (len(args) > 1 and not isinstance(args[0], int)):
            if self._filter:
                self._filter = ""
                await self.first()
            f = self._gen_filter(*args, **kwargs)
            position = await self._db.connection.get_position(self._table, ','.join(self._key_fields), f)
            if position:
                position -= 1
            else:
                position = 0
        else:
            position = args[0]
        """устанавливает курсор на заданную строку"""
        if position == 0:
            await self.first()
            return True
        elif position < 0:
            cnt = await sql(f"select count(*) from {self._table} {self._filter}", self._db, ONE)
            position = cnt + position
            if position < 0:
                position = 0
        else:
            cnt = None
        need_ofs = (position // self._page_size) * self._page_size
        need_ptr = position - need_ofs
        if need_ofs == self._offset:
            self._ptr = need_ptr
            if cnt is not None and self._auto_close_conn:
                await self._db.connection.rollback()
            return True
        if self._page:
            await self.save()
        self._page = await self(Table._FILTER, PAGE(self._page_size + 1, need_ofs))
        if not self._page:
            await self.last()
            return False
        self._ptr = need_ptr
        self._offset = need_ofs
        return True

    async def next(self):
        """
        переход на следующую запись
        если таблица еще не читалась - читает ее и встает на первую запись
        """

        if self.EOF:
            return False
        p = self.position + 1
        await self.seek(p)
        if p > self.position:
            self._ptr += 1
        return not self.EOF

    async def prior(self):
        """
        переход на предыдущую запись
        """
        p = self.position
        if p <= 0:
            return False
        await self.seek(p-1)
        return not self.BOF

    async def first(self):
        """
        переход на первую запись курсора
        """
        self._page = await self(Table._FILTER, PAGE(self._page_size + 1, 0))
        self._ptr = self._offset = 0
        return not self.EOF

    async def last(self):
        """
        переход на последнюю запись
        """
        await self.save()
        cnt = await sql(f"select count(*) from {self._table} {self._filter}", self._db, ONE)
        self._offset = cnt - self._page_size
        if self._offset < 0:
            self._offset = 0
        self._page = await self(Table._FILTER, PAGE(self._page_size+1, self._offset))
        self._ptr = len(self._page) - 1
        return not self.BOF

    @property
    def BOF(self):
        """
        :return: True, если курсор на первой записи
        """
        return (self._ptr or 0) == 0 and self._offset == 0

    @property
    def EOF(self):
        """
        :return: True, если курсор на последней записи
        """
        if self._page is None:
            return False
        p = len(self._page)
        return self._ptr >= p and p <= self._page_size

    def __aiter__(self):
        self._ptr = None
        return self

    async def __anext__(self):
        await asyncio.sleep(self._delay)
        if self._ptr is None:
            await self.first()
        else:
            await self.next()
        if self.EOF:
            raise StopAsyncIteration
        return self.row


    def __getitem__(self, item):
        return self.row[item].value

    def __setitem__(self, key, value):
        f = self.row[key]
        f.value = value

    def _gen_filter(self, *args, **kwargs)->str:

        def gen_op(lst, op):
            ret = ""
            for x in lst:
                if ret:
                    ret += f" {op} "
                if isinstance(x, Field._Condition):
                    v = x()
                elif isinstance(x, list):
                    v = gen_op(x, "or")
                elif isinstance(x, tuple):
                    v = gen_op(x,"and")
                else:
                    raise_error("BAD_V_COND", v=x)
                v = v.strip()
                if v=="":
                    raise_error("EXPR_ERROR", x=v)
                ret += f"({v})"
            return ret

        qry = gen_op(args, "and")
        nms = self.field_names
        for x in kwargs:
            if x not in nms:
                raise_error("NA_FIELD", field=x, table=self._table)
            if qry:
                qry += "and"
            qry += f"{x} = {Field._Condition._lquot(self.row._data[x]._p_type, kwargs[x])}"
        return qry

    async def filter(self, *args, **kwargs):
        """
        Задает фильтр для таблицы
        в args можно перечислять операции сравнения, например t.row.v == 1, где t - объект таблицы, м - имя поля
        операции сравнения в args могут объединяться в tuple и тогда относятся друг к другу по AND, либо в list - тогда OR
        args это tuple, соответственно соединение AND
        например, filter(t.row.a==1,[t.row.b==2,t.row.b==3]) будет преобразовано в: t.a = 1 and (t.b = 2 or t.b = 3)
        list и tuple могут быть вложены в другие list и tuple

        В kwargs все проще: ключ это имя поля, а значение - значение поля. При этом условия связаны по AND
        например, filter(a=1, b=2) будет преобразовано в: t.a = 1 and t.b = 2
        """
        if not args and not kwargs:
            self._filter = ""
        else:
            self._filter = "where " + self._gen_filter(*args, **kwargs)
        await self.first()

    async def save(self, *args, **kwargs):
        """
        При вызове без параметров, сохраняет текущие изменения в таблице
        Если в args[0] передан список записей - сохраняет его
        Если в kwargs переданы значения полей - сохраняет запись
        ---
        В случае если в таблице уже есть запись с таким ключем, как в записи, переданной в save
        происходит обновление записи, иначе вставка.
        """
        if not args and not kwargs:
            # save без аргументов = запись изменений -----------------------------------------------------------
            if self._changed:
                await self.save([self._page[x] for x in self._changed])
                self._changed.clear()
            return
        elif args:
            # список записей в первом args = сохранение каждой записи ------------------------------------------
            auto = self._auto_close_conn
            try:
                self._auto_close_conn = False
                for x in args[0]:
                    await self.save(**x)
            finally:
                self._auto_close_conn = auto
            if auto:
                await self._db.connection.commit()
            return
        #### запись **kwargs -----------------------------------------------------------------------------------
        key = ""
        if self._key_fields:
            chk = {x:kwargs[x] for x in self._key_fields if x in kwargs}
            if len(chk)==len(self._key_fields):
                for x in chk:
                    if key:
                        key += " and "
                    key += f"{self.row[x]()} = {self._db.connection.var_to_sql(chk[x])}"
        qry = ""
        if key:
            chk = await sql(f"select 1 from {self._table} where {key}", self._db, ONE)
            if chk:
                # запись найдена - обновление
                for x in kwargs:
                    if x in self._key_fields:
                        continue
                    if qry:
                        qry += ",\n"
                    qry += f"{self.row[x]()} = {self._db.connection.var_to_sql(kwargs[x])}"
                qry = f"update {self.name} set\n {qry}\n where {key}"
        if qry=="":
            # запись не найдена - вставка
            key = ""
            for x in kwargs:
                if key:
                    key += ", "
                    qry += ", "
                key += self.row[x]()
                qry += self._db.connection.var_to_sql(kwargs[x])
            qry = f"insert into {self.name}({key})\nvalues({qry})"
        await sql(qry, self._db)
        if self._auto_close_conn:
            await self._db.connection.commit()

    async def delete(self, *args, **kwargs):
        """
        При вызове без параметров удаляет строку, на которой стоит курсор
        Если вызван с параметрами - удаляет записи соответствующие фильтру, заданному параметрами
        """
        if args or kwargs:
            qry = self._gen_filter(*args, **kwargs)
        elif self.key:
            qry = {x:self.row[x].value for x in self.key}
        else:
            qry = {x:self.row[x].value for x in self.row._data.keys()}
        if isinstance(qry, dict):
            flds, qry = qry, ""
            for x in flds:
                if qry:
                    qry += " and "
                qry += f"{self.row[x]()} = {self._db.connection.var_to_sql(self.row[x].value)}"
        qry = f"delete from {self.name} where {qry}"
        await sql(qry, self._db)
        if self._auto_close_conn:
            await self._db.connection.commit()

    async def count(self, *args, **kwargs):
        """
        :param args:  --/- Если заданы, то по ним строится фильтр (аналогично filter). Иначе используется фильтр таблицы
        :param kwargs: /   Если в args передать опцию ALL, то результат будет вычислен по всей таблице (без фильтра)
        :return: Количество записей в таблице
        """
        fltr, n = "", 0
        while n < len(args):
            if args[n]==ALL:
                fltr = None
                del args[n]
            else:
                n += 1
        if fltr=="":
            fltr = "where " + self._gen_filter(*args, **kwargs) if args or kwargs else self._filter
        else:
            fltr = ""
        cnt = await sql(f"select count(*) from {self._table} {fltr}", self._db, ONE)
        if self._auto_close_conn:
            await self._db.connection.rollback()
        return cnt

    async def __call__(self, *args, **kwargs):
        """
        Возвращает данные таблицы как список строк
        умеет модифицировать результат посредством опций ONE, ROW, DICT, TUPLE, OBJECT и PAGE
        кроме того, в args и kwargs можно передать условия фильтра (аналогично filter)
        Если в args передать опцию ALL, то результат будет вычислен по всей таблице (без фильтра)
        """
        args = list(args)
        n, opts, fltr = 0, [], ""
        while n < len(args):
            if args[n] == Table._FILTER and fltr=="":
                fltr = self._filter
                del args[n]
            elif args[n] == ALL:
                fltr = None
                del args[n]
            elif isinstance(args[n], PAGE) or args[n] in (ONE, ROW, DICT, TUPLE, OBJECT):
                opts.append(args[n])
                del args[n]
            else:
                n += 1
        if (args or kwargs) and fltr is not None:
            fltr = "where " + self._gen_filter(*args, **kwargs)
        if fltr is None:
            fltr = ""
        args = opts
        args.append(self._db)
        ret = await sql(f"select * from {self._table} {fltr} order by {self._key}", *args)

        if self._auto_close_conn:
            await self._db.connection.rollback()
        return ret


class EnvTable(Table):

    @classmethod
    async def create(cls, table_name, env, page_size=100, async_delay=0.0001):
        table_struct = await env.sql("\d "+table_name)
        if not table_struct:
            raise_error("NA_TABLE", name=table_name)
        t = cls(table_struct, env, page_size, async_delay)
        await t.first()
        return t

    def __init__(self, table_struct: dict, env, page_size=100, async_delay=0.001):
        super().__init__(table_struct, env._context, page_size, False, async_delay)
        self._env = env

    def _field_changed(self, ptr, field):
        super()._field_changed(ptr, field)
        self._env.on_commit(self.save)

    async def save(self, *args, **kwargs):
        if len(args)>0 and args[0]==self._env:
            args = args[1:]
        await super().save(*args, **kwargs)
        if not args and not kwargs:
            self._env.on_commit(self.save, False)



class ClassPropertyDescriptor(object):

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self

def classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)

def find_subclass(base, name):
    if base.__name__ == name:
        return base
    for cls in base.__subclasses__():
        cls = find_subclass(cls, name)
        if cls:
            return cls
    return None


class ReadOnlyPropDict:

    def __init__(self,  **kwargs):
        self._data = kwargs
        for x in kwargs:
            self._set_f(x)

    def _set_f(self, x):
        setattr(self.__class__, x, property(lambda p: self._data[x]))

    def __getitem__(self, item):
        return self._data[item]


class PropDict(ReadOnlyPropDict):

    def _set_f(self, x):
        def set_f(p, v):
            self._data[x] = v
        setattr(self.__class__, x, property(lambda p: self._data[x], set_f))


from typing import Optional, Callable
from os import sep as file_sep
from .fnc import *
from ..error import raise_error
from ..env import Config, DB_ENV
from sys import argv as sys_argv
from pathlib import Path


def _split_module_fnc(module_name: str):
    # можно передать в пути реальное имя функции после :
    module, function = f":{module_name}".rsplit(":", 1)
    if not module:
        return function.strip(), None
    else:
        return module[1:].strip(), function.strip()


def _resolve_path(path:str):
    if path.startswith(f".{file_sep}"):
        path = Path(sys_argv[0]).parent / path[len(file_sep)+1:]
    elif path.startswith(f"..{file_sep}"):
        p = Path(sys_argv[0]).parent
        path = path[len(file_sep) + 2:]
        while path.startswith(f"..{file_sep}"):
            p = p.parent
            path = path[len(file_sep) + 2:]
        path = p.parent / path[len(file_sep) + 1:]
    return path


def _load_by_default_path(name, defs, help):
    module, function = _split_module_fnc(defs)
    if module.startswith(f".{file_sep}") or module.startswith(f"..{file_sep}"):
        module = _resolve_path(module)
    else:
        cfg = Config()
        if cfg.hasattr("defaults"):
            paths = cfg.defaults.get("plugin_path")
        else:
            paths = None
        if not paths:
            raise_error("REG_BAD_PLUG_DIR_DFLT", name=name, defs=defs)
        paths = paths.split(";") ################
        ends, module = f"{file_sep}{module}".rsplit(file_sep,1)
        if not module:
            module = ends
            ends = None
        if not module.endswith(".py"):
            module = f"{module}.py"
        for path in paths:
            path = _resolve_path(path)
            if ends and not path.endswith(ends):
                continue
            path = Path(path) / module
            if path.is_file():
                PluginFnc(name, path, help, function)
                return
    raise_error("REG_PLUG_NOT_FOUND", module=module, function=function if function else name)


def register(name:Optional[str|Callable]=None,defs=None, help=None) -> Callable:
    """
    Регистрирует запрос, плагин или функцию python как метод RPC.
    Данная функция работает и как обычная функция с параметрами и как декоратор
    :param name: имя для RPC вызова функции
    :param defs: для sql и плагинов - определение функции. для функции python - ссылка на функцию.
            при использовании в качестве декоратора, в defs может быть задано имя функции для вызова RPC
            определение sql это текст запроса, в том числе с расширениями, поддерживаемыми DB_ENV, либо tuple,
            где первый элемент это запрос, а остальные параметры, передаваемые в *args при вызове запроса.
            определение плагина не включает пробельных символов и может быть задано в одном из следующих вариантов:
            * путь к модулю, содержащему функцию
            * имя модуля, содержащего функцию (будет найден по путям поиска, заданным в конфиг. defaults.plugin_path
            * путь или имя модуля:имя функции - позволяет вызывать функцию, имя которой отличается от имени при RPC
            если register вызывается как декоратор и имя декорируемой функции SQL, то эта функция немедленно вызывается,
            результатом такой функции может быть строка запроса, либо tuple, где первый элемент это запрос, а остальные
            параметры, передаваемые в *args при вызове запроса. Документация функции SQL становится
            документацией RPC функции.
    :param help: позволяет переопределить документацию функции
    """

    def wrapper(f):
        if not iscoroutinefunction(f) and name and f.__name__ == "SQL" and len(f.__code__.co_varnames)==0:
            SqlFnc(name, f, help)
        else:
            PythonFnc(name if name else f.__name__, f, help)
        return f

    def dummy():
        ...

    if callable(name):
        PythonFnc(name.__name__, name, help or defs)
        return name
    elif callable(defs):
        PythonFnc(name, defs, help)
        return defs
    elif defs is None:
        return wrapper
    elif isinstance(defs, Path):
        PluginFnc(name, defs, help)
    elif name and (isinstance(defs, tuple) or (isinstance(defs, str) and any(x.isspace() for x in defs))):
        SqlFnc(name, defs, help)
    elif name and isinstance(defs, str) and not (defs.startswith("/") or defs.startswith("~")):
        _load_by_default_path(name, defs, help)
    elif name and isinstance(defs, str):
        module, function = _split_module_fnc(defs)
        PluginFnc(name, module, help, function)
    else:
        raise_error("BAD_FN_PARAMS", name=name, defs=defs)
    return dummy


async def call_rpc(name_of_rpc_method_for_call, current_rpc_environment_for_call=None, **kwargs):
    """
    Выполняет вызов RPC
    :param current_rpc_environment_for_call: окружение, в котором будет выполняться метод
    :param name_of_rpc_method_for_call: имя метода
    :param kwargs: параметры
    :return: результат выполнения метода
    """
    if current_rpc_environment_for_call is None:
        current_rpc_environment_for_call = DB_ENV()
        local = True
    else:
        local = False
    if isinstance(name_of_rpc_method_for_call, str):
        f = Fnc.search(name_of_rpc_method_for_call)
        if f is None:
            raise_error("FN_NOT_FOUND", method=name_of_rpc_method_for_call)
    elif not isinstance(name_of_rpc_method_for_call, Fnc):
        raise_error("BAD_RPC_METHOD")
    else:
        f = name_of_rpc_method_for_call
    try:
        ret = await f(current_rpc_environment_for_call, **kwargs)
        if local and current_rpc_environment_for_call.in_transaction:
            await current_rpc_environment_for_call.commit()
        return ret
    except Exception as e:
        if current_rpc_environment_for_call.in_transaction:
            await current_rpc_environment_for_call.rollback()
        raise e


import sys, os
sys.path.append(os.path.dirname(__file__))
from driver import Driver as BaseDriver, RowFactory
from sqlite3 import connect as db_connect
from collections import namedtuple, OrderedDict


def dict_factory(cursor, row):
    fields = [column[0] for column in cursor.description]
    return {key: value for key, value in zip(fields, row)}


def namedtuple_factory(cursor, row):
    fields = [column[0] for column in cursor.description]
    cls = namedtuple("Row", fields)
    return cls._make(row)

_Info = namedtuple("Info", ["column_name", "data_type", "is_nullable", "character_maximum_length",
                       "numeric_precision", "numeric_scale", "column_default", "autoincrement", "pk"])

def _field_ifo_tuples(fields, str_fields):
    str_fields = str_fields["sql"].split("\n", 1)[1][:-1].strip().split("\n")
    str_def = {}
    for x in str_fields:
        n, x = x.strip().split(" ", 1)
        str_def[n] = x
    for n, row in enumerate(fields):
        if "(" in row["type"]:
            sz = row["type"].split("(")[1][:-1]
            sz = int(sz)
        else:
            sz = None
        ai = str_def[row["name"]]
        fields[n] = _Info(column_name=row["name"], data_type=row["type"].lower(), character_maximum_length=sz,
                          column_default=row["dflt_value"], is_nullable=row["notnull"] == 0 and row["pk"] == 0,
                          autoincrement=" autoincrement" in ai, pk=row["pk"],
                          numeric_precision=None, numeric_scale=None)
    return fields

_FACTORY_ = {
    RowFactory.ANY.value: None,
    RowFactory.TUPLE.value: None,
    RowFactory.DICT.value: dict_factory,
    RowFactory.NAMED_TUPLE.value: namedtuple_factory,
}

class Driver(BaseDriver):

    def __init__(self, connection_string, on_open_close=None):
        """
        test_db.sqlite
        """
        super().__init__(connection_string, on_open_close)

    async def begin(self):
        await self._before_open()
        self._conn = db_connect(self.connection_string)

    async def sql(self, query, one_row=False, row_factory=RowFactory.DICT):
        if self._conn is None:
            await self.begin()
        acur = self._conn.cursor()
        if x:=_FACTORY_[row_factory.value]:
            acur.row_factory = x
        acur.execute(query)
        descr = acur.description
        if not descr:
            return acur.rowcount
        if one_row:
            ret = acur.fetchone()
        else:
            ret = acur.fetchall()
            if not ret:
                ret = []
        return ret

    async def commit(self):
        if not self._conn:
            raise RuntimeError("Transaction is not open")
        self._conn.commit()
        await self.rollback()

    async def rollback(self):
        if not self._conn:
            return
        self._conn.close()
        await super().rollback()

    async def table_struct(self, table_name) -> dict:
        defs = await self.sql(f"select sql from sqlite_master where type='table' and name='{table_name}'", one_row=True)
        if not defs:
            return None
        fields = await self.sql(f"PRAGMA table_info('{table_name}')")
        fields, pk, defs = [], [], _field_ifo_tuples(fields, defs)
        for x in defs:
            t = self._specify_type(x)
            field = {
                "name": x.column_name,
                "type": t.type_name,
                "not_null": x.is_nullable is not None and f'{x.is_nullable} '[0].upper() in "FNН",
                "python_type": t.python_type,
            }
            if t.autoincrement or x.autoincrement:
                field["autoincrement"] = True
            n = self.ident_name(field["name"])
            if n != field["name"]:
                field["field"] = n
            if x.character_maximum_length:
                field["size"] = x.character_maximum_length
            elif field["python_type"] != int and x.numeric_precision:
                n = f'{x.numeric_precision}.{x.numeric_scale}'
                field["size"] = float(n)
            if x.pk:
                while len(pk) < x.pk:
                    pk.append(None)
                pk[x.pk-1] = x.column_name
            fields.append(field)
        constraints = []
        if pk:
            constraints.append({
                "type": "primary key",
                "fields": pk
            })
        defs = await self.sql(f"PRAGMA foreign_key_list('{table_name}')")
        fk = {}
        for x in defs:
            k = fk.get(x["id"], {"type": "foreign key", "table":x["table"], "fields":OrderedDict()})
            k["fields"][x["from"]] = x["to"]
            fk[x["id"]] = k
        for x in fk:
            constraints.append(fk[x])

        defs = await self.sql(f"select sql from sqlite_master where type = 'index'  and sql like '%UNIQUE INDEX % on {table_name} %'",
                              row_factory=RowFactory.TUPLE)
        for x in defs:
            x = x[0].split("(",1)[1][:-1]
            constraints.append({
                "type": "unique",
                "fields": [f.strip() for f in x.split(",")],
            })
        return {
            "table": table_name,
            "fields": fields,
            "constraints": constraints
        }


from abc import ABC
from ..class_tools import classproperty
from .driver import RowFactory, Driver, _set_is_option
from pathlib import Path
from importlib.machinery import SourceFileLoader
from json5 import loads
from asyncio import BoundedSemaphore, wait_for, TimeoutError, sleep
from ..error import raise_error
from inspect import iscoroutinefunction
from typing import Callable
from gc import collect as garbage_collect

DRIVER_CLASSES = {}
_DRIVERS_ = {}
_LIMITS_ = {}
_CFG_ = None

def _check_cfg(key=""):
    global _CFG_
    if _CFG_ is None:
        from ..env import Config
        _CFG_ = Config()
    if key=="":
        return
    if _CFG_.hasattr("defaults"):
        return _CFG_.defaults.get(key)


class Option(ABC):

    @staticmethod
    def is_option(x):
        if isinstance(x, Option):
            return True
        while hasattr(x, "__base__"):
            if x.__base__ == Option:
                return True
            x = x.__base__
        return False

    @classproperty
    def row_factory(cls)->RowFactory:
        return RowFactory.ANY

    @classproperty
    def can_process(cls):
        return False

    @classproperty
    def one_row(cls):
        return None

    @staticmethod
    async def process(ret_data, connection, row_factory):
        """
        Возвращает преобразованный набор данных и формат
        если вместо формата вернулся None - дальнейшее преобразование невозможно
        """
        return ret_data, row_factory

_set_is_option(Option.is_option)

class ALL(Option):
    ...

class DB(Option):
    """
    Соединение с БД
    """
    _TRASH_ = set()

    def __init__(self, connection_string: str=""):
        connection_string = connection_string.strip()
        if connection_string.startswith("jdbc:"):
            connection_string = connection_string[5:]
        if connection_string.endswith("}"):
            connection_string, params = connection_string.split("{",1)
            self._params = loads(f"{{{params}")
            for x in tuple(self._params.keys()):
                if "$" in x:
                    v = self._params[x]
                    del self._params[x]
                    self._params[x.replace("$",".")] = v
        else:
            self._params = {}
        if "://" not in connection_string:
            _check_cfg()
            connection_string = _CFG_.db_connection(connection_string)
        driver_name, connection_string = connection_string.split("://", 1)

        connection_string, conn_params = f"{connection_string}?".split("?",1)

        driver = DRIVER_CLASSES.get(driver_name)
        driver_path = None
        if conn_params:
            conn_params = conn_params[:-1].split("&")
            if "driver_path" in conn_params:
                driver_path = conn_params["driver_path"]
                del conn_params["driver_path"]
            conn_params = "?" + "&".join(conn_params)
        connection_string += conn_params
        if driver is None:
            if driver_path is None:
                driver_path = _check_cfg("db_driver_path")
            if driver_path is None:
                driver_path = Path(__file__).parent / f"driver_{driver_name}.py"
            elif isinstance(driver_path, str):
                driver_path = Path(driver_path) / f"driver_{driver_name}.py"
            m = driver_path.name
            driver_path = str(driver_path)
            if driver_path in _DRIVERS_:
                driver = _DRIVERS_[driver_path]
            else:
                m = m.split(".",1)[0] + f"_{len(_DRIVERS_)}"
                m = SourceFileLoader(m, driver_path ).load_module()
                driver = m.Driver
                _DRIVERS_[driver_path] = driver
        else:
            driver_path = ""

        self._hash = hash(driver_path + connection_string)
        if "LIMIT" in self._params:
            m = self._params["LIMIT"]
            del self._params["LIMIT"]
        else:
            m = 0
        self._conn_limit = None
        self.connection_limit = m
        self._connection = driver(connection_string, self._on_open_close)

    @property
    def connection_limit(self)->int:
        lmt = _LIMITS_.get(self._hash)
        return lmt._value if lmt else 0


    @connection_limit.setter
    def connection_limit(self, value:int):
        """
        Позволяет задать ограничение на количество таких соединений
        """
        if value<=0:
            if self._hash in _LIMITS_:
                del _LIMITS_[self._hash]
            return
        elif lmt:=_LIMITS_.get(self._hash):
            if lmt._waiters:
                lmt._value = value - len(lmt._waiters)
            else:
                lmt._value = value
        else:
            lmt = BoundedSemaphore(value)
            _LIMITS_[self._hash] = lmt
        self._conn_limit = lmt


    async def _on_open_close(self, close=False)->dict:
        if self._conn_limit:
            if close:
                await self._conn_limit.release()
                if self._hash not in _LIMITS_:
                    self._conn_limit = None
            elif self._hash not in _LIMITS_:
                self._conn_limit = None
            else:
                if self._conn_limit.locked():
                    await self.garbage_collect()
                await self._conn_limit.acquire()
        return self._params

    @property
    def connection(self)->Driver:
        return self._connection

    def __del__(self):
        if self.connection is not None and self.connection.in_transaction:
            DB._TRASH_.add(self.connection)

    @classmethod
    async def garbage_collect(cls, clear_mem=True):
        if clear_mem:
            garbage_collect()
        while len(DB._TRASH_)>0:
            x = DB._TRASH_.pop()
            await x.rollback()

class TIMEOUT(Option):

    def __init__(self, time_sec:float, raise_error:bool=True):
        self._time_sec = time_sec
        self._raise_error = raise_error

    async def __call__(self, func):
        try:
            return await wait_for(func, self._time_sec)
        except TimeoutError as e:
            if self._raise_error:
                raise e
            return None

class DICT(Option):
    """
    Указывает, что результат следует вернуть как список dict (по умолчанию)
    """
    @classproperty
    def row_factory(cls) -> RowFactory:
        return RowFactory.DICT

class TUPLE(Option):
    """
    Указывает, что результат следует вернуть как список tuple
    """
    @classproperty
    def row_factory(cls) -> RowFactory:
        return RowFactory.TUPLE

class OBJECT(Option):
    """
    Указывает, что результат следует вернуть как список namedtuple
    """
    @classproperty
    def row_factory(cls) -> RowFactory:
        return RowFactory.NAMED_TUPLE

class ROW(Option):
    """
    Указывает, что нужно вернуть только первую строку из выборки
    """
    @classproperty
    def one_row(cls):
        return True

class ONE(ROW):
    """
    Указывает, что нужно вернуть только значение первого поля из первой строки выборки
    """
    @classproperty
    def can_process(cls):
        return True

    @staticmethod
    async def process(ret_data, connection, row_factory):
        if isinstance(ret_data, (int, None.__class__)):
            return ret_data, None
        match row_factory:
            case RowFactory.TUPLE:
                return ret_data[0], None
            case RowFactory.DICT:
                return ret_data[tuple(ret_data.keys())[0]], None
            case RowFactory.NAMED_TUPLE:
                x = ret_data._fields
                if len(x) < 1:
                    return None, None
                x = x[0]
                return getattr(ret_data, x), None
            case _:
                return ret_data, None

class JSON(ONE):
    """
    Указывает, что значение первого поля в первой строке следует привести к dict или list
    """
    @staticmethod
    async def process(ret_data, connection, row_factory):
        ret_data, _ = await ONE.process(ret_data, connection, row_factory)
        if isinstance(ret_data, (dict, list)):
            return ret_data, None
        elif isinstance(ret_data, str):
            return loads(ret_data), None
        else:
            raise_error("NOT_CONV_DICT_LIST", data=ret_data)

class ROLLBACK(Option):
    """
    Указывает, что после выполнения запроса соединение должно быть закрыто с откатом транзакции
    """
    @classproperty
    def can_process(cls):
        return True

    @staticmethod
    async def process(ret_data, connection, row_factory):
        if connection.in_transaction:
            await connection.rollback()
        return ret_data, row_factory

class COMMIT(Option):
    """
    Указывает, что после выполнения запроса соединение должно быть закрыто с подтверждением транзакции
    """
    @classproperty
    def can_process(cls):
        return True

    @staticmethod
    async def process(ret_data, connection, row_factory):
        if connection.in_transaction:
            await connection.commit()
        return ret_data, row_factory

class PAGE(Option):

    def __init__(self, limit:int, offset:int=0):
        """
        Ограничивает количество записей на странице и задает смещение первой записи страницы от начала выборки
        :param limit: максимальное количество строк
        :param offset: смещение от начала
        """
        if not(isinstance(limit, int) and isinstance(offset, int)) or limit < 0 or offset < 0:
            raise_error("BAD_LIMIT_OFFSET")
        self._limit = limit
        self._offset = offset

    async def __call__(self, connection, query):
        query = await connection.page(query, self._limit, self._offset)
        return query

    def __str__(self):
        return f"PAGE(limit={self._limit}, offset={self._offset})"


class CALLBACK(Option):

    def __init__(self, callback_function:Callable, attribute_collection=None, *args, **kwargs):
        """
        Позволяет вернуть запрос с подставленными параметрами из функции sql и доработать его или вывести в лог
        если callback_function возвращает строку, то именно эта строка станет запросом
        :param callback_function: ссылка на функцию, которой передается текст запроса
        """
        self._callback = callback_function
        self._args = args
        self._kwargs = kwargs
        self._attrs = attribute_collection

    async def __call__(self, query):
        if not self._callback:
            return query
        if iscoroutinefunction(self._callback):
            x = await self._callback(query, *self._args, **self._kwargs)
        else:
            x = self._callback(query, *self._args, **self._kwargs)
        return x or query

    def __getitem__(self, item):
        if self._attrs is None:
            raise_error("ATTR_NOT_FOUND", name=item)
        return self._attrs[item]


class ITERATOR(Option):

    def __init__(self, page_size:int=200, async_delay=0.00000001):
        self._page_size = page_size
        self._delay = async_delay
        self._ofs = 0
        self._pos = page_size-1
        self._buf = None
        self._query = None
        self._db = None
        self._row_factory = None
        self._process = None

    def __call__(self, query, db=None, row_factory=RowFactory.DICT, process=None):
        if self._query:
            return None
        self._query = query
        self._db = db
        self._row_factory = row_factory
        self._process = process if isinstance(process, list) else []
        return self

    def __aiter__(self):
        self._ofs = -self._page_size
        self._pos = self._page_size - 1
        self._buf = None
        return self

    async def __anext__(self):
        await sleep(self._delay)
        self._pos += 1
        if self._buf and self._pos >= len(self._buf) and len(self._buf) < self._page_size:
            raise StopAsyncIteration
        elif self._pos >= self._page_size:
            self._ofs += self._page_size
            self._pos = 0
            q = await self._db.connection.page(self._query, self._page_size, self._ofs)
            self._buf = await self._db.connection.sql(q, one_row=False, row_factory=self._row_factory)
            if not self._buf:
                raise StopAsyncIteration
        return self._buf[self._pos]

